/* xgetopt.c - Similar behaviour to getopt and getopt_long, but not an
 *             attempt to emulate those exactly.
 *
 * Copyright (C) 2013, 2020 Timothy Richard Musson
 *
 * Email: Tim Musson <trmusson@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *
 * gcc -D TEST xgetopt.c -o xgetopt-test
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>

#ifndef XUTIL_H
#include "xutil.h"
#endif

#ifndef XGETOPT_H
#include "xgetopt.h"
#endif



int
xgetopt (int argc, char *argv[], int *idx, XGetOption *option, int n_options)
{
	int opt, len;
	xbool_t not_last;
	const char *str;


	str = argv[*idx];

	if (*str != '-') {
		return XGETOPT_NONE;
	}

	*idx = *idx + 1;

	str++;
	if (*str == '-') {
		str++;
	}
	len = strlen (str);

	not_last = *idx < argc;

#if XGETOPT_SENSITIVE

	/* In this version, short and long forms are fully case-sensitive. */

	if (len > 0) {
		for (opt = 0; opt < n_options; opt++) {
			if (len >= option[opt].min_len &&
				strncmp (str, option[opt].opt_str, len) == 0) {
				if (!option[opt].requires_arg || not_last) {
					return opt;
				}
				return XGETOPT_MISSING_ARG;
			}
		}
	}

#else

	/* In this version, the first character is case-sensitive,
	   but only when len and min_len are both 1. For example, this
	   allows a program to differentiate between "-w", "--width",
	   "-W", and "--window" by using "width" and "Window" as the
	   option strings.  */

	if (len > 0) {
		xbool_t match;
		for (opt = 0; opt < n_options; opt++) {
			if (len >= option[opt].min_len_short) {

				/* Match the 1st char. */
				match = (len == 1 && option[opt].min_len_short == 1 ?
				         *str == *(option[opt].opt_str)
				         : strncasecmp (str, option[opt].opt_str, 1) == 0);

				/* If it's a long form, ensure it's no shorter than
				   min_len_long. */
				match = (match && len > option[opt].min_len_short ?
				         len >= option[opt].min_len_long : match);

				/* Match the rest. */
				if (match && strncasecmp (str + 1, option[opt].opt_str + 1,
				                          len - 1) == 0) {
					if (!option[opt].requires_arg || not_last) {
						return opt;
					}
					return XGETOPT_MISSING_ARG;
				}
			}
		}
	}

#endif

	return XGETOPT_UNKNOWN;
}


#ifdef TEST

static XGetOption option[] = {
#if XGETOPT_SENSITIVE
	{1, TRUE,  "foo"},
	{1, TRUE,  "bar"},
	{1, TRUE,  "Baz"},
	{3, FALSE, "width"},
	{3, FALSE, "window"},
	{1, FALSE, "help"}
#else
	{1, 1, TRUE,  "foo"},
	{1, 1, TRUE,  "bar"},
	{1, 1, TRUE,  "Baz"},
	{1, 3, FALSE, "width"},
	{1, 3, FALSE, "Window"},
	{1, 1, FALSE, "help"}
#endif
};

enum {
	OPT_FOO,
	OPT_BAR,
	OPT_BAZ,
	OPT_WIDTH,
	OPT_WINDOW,
	OPT_HELP,
	N_OPTS
};


static void
show_usage (const char *progname)
{
	printf ("\nUsage: %s [OPTIONS] ...\n\n", progname);
	printf ("Options:\n\n");
	printf ("  -f,   --foo VALUE   Give a value for foo\n");
#if XGETOPT_SENSITIVE
	printf ("  -b,   --bar VALUE   Give a value for bar\n");
	printf ("  -B,   --Baz VALUE   Give a value for baz\n");
	printf ("  -wid, --width       Make width true\n");
	printf ("  -win, --window      Make window true\n");
#else
	printf ("  -b,   --bar VALUE   Give a value for bar\n");
	printf ("  -B,   --baz VALUE   Give a value for baz\n");
	printf ("  -w,   --width       Make width true\n");
	printf ("  -W,   --window      Make window true\n");
#endif
	printf ("  -h,   --help        Show this usage information\n");
	printf ("\n");
}


int
main (int argc, char *argv[])
{
	int opt, idx;
	const char *foo, *bar, *baz;
	xbool_t width, window;

	foo = bar = baz = "";
	width = window = FALSE;

	opt = 1;
	idx = 1;

	while (idx < argc && opt > XGETOPT_NONE) {

		opt = xgetopt (argc, argv, &idx, option, N_OPTS);

		switch (opt) {

		case OPT_HELP:
			show_usage (argv[0]);
			exit (0);
			break;

		case OPT_FOO:
			foo = argv[idx++];
			break;

		case OPT_BAR:
			bar = argv[idx++];
			break;

		case OPT_BAZ:
			baz = argv[idx++];
			break;

		case OPT_WIDTH:
			width = TRUE;
			break;

		case OPT_WINDOW:
			window = TRUE;
			break;

		case XGETOPT_MISSING_ARG:
			fprintf (stderr, "option '%s' requires an argument\n",
			         argv[idx - 1]);
			exit (1);

		case XGETOPT_UNKNOWN:
			fprintf (stderr, "unknown option '%s'\n", argv[idx - 1]);
			exit (1);

		default:
			/* assert - shouldn't get here */
			break;

		}

	}

	printf ("width = %d, window = %d, foo = %s, bar = %s, baz = %s, idx = %d, argc = %d\n",
	        width, window, foo, bar, baz, idx, argc);

	while (idx < argc) {
		printf ("%s\n", argv[idx]);
		idx++;
	}

	return 0;
}

#endif

